# SPDX-License-Identifier: Apache-2.0

import collections

from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_matmul, apply_exp, apply_add
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers


def convert_aft_survival_regression(scope, operator, container):
    op = operator.raw_operator

    coefficients = op.coefficients.toArray().astype(float)
    coefficients_tensor = scope.get_unique_variable_name('coefficients_tensor')
    container.add_initializer(coefficients_tensor, onnx_proto.TensorProto.FLOAT, [1, len(coefficients)], coefficients)
    intercepts = (
        op.intercept.astype(float)
        if isinstance(op.intercept, collections.abc.Iterable)
        else [float(op.intercept)])
    intercepts_tensor = scope.get_unique_variable_name('intercepts_tensor')
    container.add_initializer(intercepts_tensor, onnx_proto.TensorProto.FLOAT, [len(intercepts)], intercepts)

    matmul_result = scope.get_unique_variable_name('matmul_result_tensor')
    apply_matmul(scope, [operator.input_full_names[0], coefficients_tensor], matmul_result, container)
    add_result = scope.get_unique_variable_name('intercept_added_tensor')
    apply_add(scope, [matmul_result, intercepts_tensor], add_result, container)
    apply_exp(scope, add_result, operator.output_full_names, container)


register_converter('pyspark.ml.regression.AFTSurvivalRegressionModel', convert_aft_survival_regression)


def calculate_aft_survival_regression_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, 1])


register_shape_calculator('pyspark.ml.regression.AFTSurvivalRegressionModel',
                          calculate_aft_survival_regression_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import unittest
from onnxmltools.convert.sparkml.operator_converters.tree_helper import Node


class TestSparkmDecisionTreeClassifier(unittest.TestCase):

    def test_rule_number(self):
        attrs = {
            'class_ids': [0, 0, 0, 0, 0, 0],
            'class_nodeids': [2, 4, 5, 8, 9, 10],
            'class_treeids': [0, 0, 0, 0, 0, 0],
            'class_weights': [0.8462194428652643,
                           0.2781875658587987,
                           0.5437174290677474,
                           0.6656197654941374,
                           0.4343004513217279,
                           0.2975769813225644],
            'classlabels_int64s': [0, 1],
            'nodes_falsenodeids': [6, 3, 0, 5, 0, 0, 10, 9, 0, 0, 0],
            'nodes_featureids': [0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False],
            'nodes_modes': ['BRANCH_LEQ',
                         'BRANCH_LEQ',
                         'LEAF',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'BRANCH_LEQ',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'LEAF'],
            'nodes_nodeids': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [1, 2, 0, 4, 0, 0, 7, 8, 0, 0, 0],
            'nodes_values': [21.65000057220459,
                          100.95000076293945,
                          0.0,
                          -22.84999942779541,
                          0.0,
                          0.0,
                          98.0999984741211,
                          37.14999961853027,
                          0.0,
                          0.0,
                          0.0],
            'post_transform': 'NONE'}
        root, _ = Node.create(attrs)
        root.unfold_rule_or()
        new_attrs = root.to_attrs(
                post_transform=attrs['post_transform'],
                classlabels_int64s=attrs["classlabels_int64s"])
        assert len(attrs['nodes_nodeids']) <= len(new_attrs['nodes_nodeids'])
        for i in range(len(new_attrs["nodes_truenodeids"])):
            if new_attrs["nodes_modes"][i] == 'LEAF':
                continue
            assert new_attrs["nodes_truenodeids"][i] > i
        for i in range(len(new_attrs["nodes_falsenodeids"])):
            if new_attrs["nodes_modes"][i] == 'LEAF':
                continue
            assert new_attrs["nodes_falsenodeids"][i] > i

        attrs = {
            'class_ids': [0, 0, 0, 0, 0, 0],
            'class_nodeids': [102, 104, 105, 108, 109, 1010],
            'class_treeids': [0, 0, 0, 0, 0, 0],
            'class_weights': [0.8462194428652643,
                           0.2781875658587987,
                           0.5437174290677474,
                           0.6656197654941374,
                           0.4343004513217279,
                           0.2975769813225644],
            'classlabels_int64s': [0, 1],
            'nodes_falsenodeids': [106, 103, 0, 105, 0, 0, 1010, 109, 0, 0, 0],
            'nodes_featureids': [0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False],
            'nodes_modes': ['BRANCH_LEQ',
                         'BRANCH_LEQ',
                         'LEAF',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'BRANCH_LEQ',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'LEAF'],
            'nodes_nodeids': [100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 1010],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [101, 102, 0, 104, 0, 0, 107, 108, 0, 0, 0],
            'nodes_values': [21.65000057220459,
                          100.95000076293945,
                          0.0,
                          -22.84999942779541,
                          0.0,
                          0.0,
                          98.0999984741211,
                          37.14999961853027,
                          0.0,
                          0.0,
                          0.0],
            'post_transform': 'NONE'}
        root, _ = Node.create(attrs)
        root.unfold_rule_or()
        new_attrs2 = root.to_attrs(
                post_transform=attrs['post_transform'],
                classlabels_int64s=attrs["classlabels_int64s"])
        assert len(attrs['nodes_nodeids']) <= len(new_attrs2['nodes_nodeids'])
        for k in new_attrs:
            if new_attrs[k] != new_attrs2[k]:
                raise ValueError(f"Key {k!r}, {new_attrs[k]} != {new_attrs2[k]}")
        self.assertEqual(new_attrs, new_attrs2)

    def test_rule_in_set(self):
        attrs = {
            'class_ids': [0, 0, 0, 0, 0, 0],
            'class_nodeids': [2, 4, 5, 8, 9, 10],
            'class_treeids': [0, 0, 0, 0, 0, 0],
            'class_weights': [0.8462194428652643,
                           0.2781875658587987,
                           0.5437174290677474,
                           0.6656197654941374,
                           0.4343004513217279,
                           0.2975769813225644],
            'classlabels_int64s': [0, 1],
            'nodes_falsenodeids': [6, 3, 0, 5, 0, 0, 10, 9, 0, 0, 0],
            'nodes_featureids': [0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False,
                                             False],
            'nodes_modes': ['||',
                         'BRANCH_LEQ',
                         'LEAF',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'BRANCH_LEQ',
                         'BRANCH_LEQ',
                         'LEAF',
                         'LEAF',
                         'LEAF'],
            'nodes_nodeids': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [1, 2, 0, 4, 0, 0, 7, 8, 0, 0, 0],
            'nodes_values': [[21.56, 22.56],
                          100.95000076293945,
                          0.0,
                          -22.84999942779541,
                          0.0,
                          0.0,
                          98.0999984741211,
                          37.14999961853027,
                          0.0,
                          0.0,
                          0.0],
            'post_transform': 'NONE'}
        root, _ = Node.create(attrs)
        root.unfold_rule_or()
        new_attrs = root.to_attrs(
                post_transform=attrs['post_transform'],
                classlabels_int64s=attrs["classlabels_int64s"])
        assert len(attrs['nodes_nodeids']) < len(new_attrs['nodes_nodeids'])
        assert "BRANCH_EQ" in new_attrs["nodes_modes"]
        assert "||" not in new_attrs["nodes_modes"]


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import CountVectorizer, CountVectorizerModel
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase

TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())

class TestSparkmlCountVectorizer(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_count_vectorizer_default(self):
        data = self.spark.createDataFrame([
            ("A B C".split(" "), ),
            ("A B B C A".split(" "), ),
        ], ["text"])
        count_vec = CountVectorizer(inputCol="text", outputCol="result", minTF=1.0, binary=False)
        model: CountVectorizerModel = count_vec.fit(data)
        result = model.transform(data)

        model_onnx = convert_sparkml(model, 'Sparkml CountVectorizer', [('text', StringTensorType([None, None]))], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        
        data_pd = data.toPandas()
        data_np = {
            "text": data_pd.text.apply(lambda x: pandas.Series(x)).values.astype(str),
        }

        expected = {
            "prediction_result": numpy.asarray(result.toPandas().result.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)),
        }

        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlCountVectorizerModel_Default")
        onnx_model_path = paths[-1]
        
        output_names = ['result']
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))
        
        assert output_shapes[0] == [None, 3]
        compare_results(expected["prediction_result"], actual_output["result"], decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_count_vectorizer_binary(self):
        data = self.spark.createDataFrame([
            ("A B C".split(" "), ),
            ("A B B C A".split(" "), ),
            ("B B B D".split(" "), ),
        ], ["text"])
        count_vec = CountVectorizer(inputCol="text", outputCol="result", minTF=2.0, binary=True)
        model: CountVectorizerModel = count_vec.fit(data)
        result = model.transform(data)

        model_onnx = convert_sparkml(model, 'Sparkml CountVectorizer', [('text', StringTensorType([None, None]))], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        
        data_pd = data.toPandas()
        data_np = {
            "text": data_pd.text.apply(lambda x: pandas.Series(x)).values.astype(str),
        }

        expected = {
            "prediction_result": numpy.asarray(result.toPandas().result.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)),
        }

        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlCountVectorizerModel_Binary")
        onnx_model_path = paths[-1]
        
        output_names = ['result']
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))
        
        assert output_shapes[0] == [None, 4]
        compare_results(expected["prediction_result"], actual_output["result"], decimal=5)

if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_batch_norm, apply_instance_norm
from ....common._registration import register_converter


def convert_batch_normalization(scope, operator, container):
    params = operator.raw_operator.batchnorm

    if params.instanceNormalization and not params.computeMeanVar:
        raise ValueError('It is impossible to do instance normalization without re-computing mean and variance')

    if params.instanceNormalization and params.computeMeanVar:
        op_type = 'InstanceNormalization'
    else:
        op_type = 'BatchNormalization'

    inputs = [operator.inputs[0].full_name]
    outputs = [operator.outputs[0].full_name]
    scale_tensor_name = scope.get_unique_variable_name(op_type + '_scale')
    container.add_initializer(scale_tensor_name, onnx_proto.TensorProto.FLOAT, [params.channels],
                              params.gamma.floatValue)
    inputs.append(scale_tensor_name)
    bias_tensor_name = scope.get_unique_variable_name(op_type + '_B')
    container.add_initializer(bias_tensor_name, onnx_proto.TensorProto.FLOAT, [params.channels], params.beta.floatValue)
    inputs.append(bias_tensor_name)

    epsilon = params.epsilon

    if op_type == 'BatchNormalization':
        mean_tensor_name = scope.get_unique_variable_name(op_type + '_mean')
        container.add_initializer(mean_tensor_name, onnx_proto.TensorProto.FLOAT, [params.channels],
                                  params.mean.floatValue)
        inputs.append(mean_tensor_name)
        variance_tensor_name = scope.get_unique_variable_name(op_type + '_variance')
        container.add_initializer(variance_tensor_name, onnx_proto.TensorProto.FLOAT, [params.channels],
                                  params.variance.floatValue)
        inputs.append(variance_tensor_name)
        momentum = 0.
        spatial = 1  # True

        if not params.instanceNormalization and params.computeMeanVar:
            # In this case, we apply batch normalization and adjust the statistics stored according the the batch
            # being processed.

            # To update "mean" and "var," we put their updated results back to the associated input tensors.
            outputs += inputs[1:3]
            # We also allocate two extra output buffers to store some intermediate results, but they are not used
            # in CoreML model.
            outputs.append(scope.get_unique_variable_name('saved_mean'))
            outputs.append(scope.get_unique_variable_name('saved_var'))
            # We choose "training" mode because some variables need to be updated.
            is_test = 0  # False
        elif not params.instanceNormalization and not params.computeMeanVar:
            # In this case, batch normalization is applied without updating mean, variance, etc. according to
            # the batches being processed. It means this operator works under testing model. Because there is no
            # variable update, we don't need to specify extra inputs and outputs like in previous code block.
            is_test = 1  # True
        else:
            raise ValueError('Unsupported operation mode')

        apply_batch_norm(scope, inputs, outputs, container, operator_name=operator.full_name, epsilon=epsilon,
                         is_test=is_test, momentum=momentum, spatial=spatial)
    else:
        apply_instance_norm(scope, inputs, outputs, container, operator_name=operator.full_name, epsilon=epsilon)


register_converter('batchnorm', convert_batch_normalization)



# SPDX-License-Identifier: Apache-2.0

"""
Helpers to help testing deep learning converted models.
"""
import numpy as np
from .main import save_model

rt_onnxruntime, rt_caffe2, rt_cntk = range(3)
_rt_installed = None


def find_inference_engine():
    """
    Finds an available backend to evaluate deep learning models.
    Returns its name as a string.
    """
    global _rt_installed
    if _rt_installed is not None:
        return _rt_installed

    try:
        import onnxruntime
        _rt_installed = rt_onnxruntime
    except ImportError:
        try:
            import cntk
            _rt_installed = rt_cntk
        except ImportError:
            try:
                import caffe2
                _rt_installed = rt_caffe2
            except ImportError:
                pass

    return _rt_installed


def evaluate_deep_model(onnx_model, inputs, rt_type=None):
    """
    Evaluates a deep learning model with
    onnxruntime, *CNTK* or *caffe2*
    """
    if rt_type is None:
        rt_type = find_inference_engine()
    if rt_type == rt_onnxruntime:
        return _evaluate_onnxruntime(onnx_model, inputs)
    if rt_type == rt_cntk:
        return _evaluate_cntk(onnx_model, inputs)
    elif rt_type == rt_caffe2:
        return _evaluate_caffe2(onnx_model, inputs)
    else:
        raise ImportError('No runtime found. Need either CNTK or Caffe2')


def _evaluate_onnxruntime(onnx_model, inputs):
    import onnxruntime
    runtime = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    result = None
    inputs = inputs if isinstance(inputs, list) else [inputs]
    for i_ in range(inputs[0].shape[0]):  # TODO: onnxruntime can't support batch_size > 1
        out = runtime.run([], {x.name: inputs[n_][i_:i_ + 1] for n_, x in enumerate(runtime.get_inputs())})[0]
        result = out if result is None else np.concatenate((result, out))

    return result[0] if isinstance(result, list) else result


def _evaluate_caffe2(onnx_model, inputs):
    from caffe2.python.onnx.backend import run_model

    if not isinstance(inputs, list):
        inputs = [inputs]

    outputs = run_model(onnx_model, inputs)

    adjusted_outputs = dict()
    for output in onnx_model.graph.output:
        adjusted_outputs[output.name] = outputs[output.name]

    return adjusted_outputs[onnx_model.graph.output[0].name]


def _evaluate_cntk(onnx_model, inputs):
    import cntk
    if not isinstance(inputs, list):
        inputs = [inputs]

    adjusted_inputs = dict()

    for i, x in enumerate(inputs):
        onnx_name = onnx_model.graph.input[i].name
        adjusted_inputs[onnx_name] = [np.ascontiguousarray(np.squeeze(_, axis=0)) for _ in np.split(x, x.shape[0])]

    temporary_onnx_model_file_name = 'temp_' + onnx_model.graph.name + '.onnx'
    save_model(onnx_model, temporary_onnx_model_file_name)
    cntk_model = cntk.Function.load(temporary_onnx_model_file_name, format=cntk.ModelFormat.ONNX)

    return cntk_model.eval(adjusted_inputs)


def create_tensor(N, C, H=None, W=None):
    if H is None and W is None:
        return np.random.rand(N, C).astype(np.float32, copy=False)
    elif H is not None and W is not None:
        return np.random.rand(N, C, H, W).astype(np.float32, copy=False)
    else:
        raise ValueError('This function only produce 2-D or 4-D tensor')



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ...common._registration import register_converter


def convert_glm_classifier(scope, operator, container):
    # For classifiers, due to the different representations of classes' probabilities in ONNX and CoreML, some extra
    # operators are required. See explanation below.
    #
    # Symbols:
    #  X: input feature vector
    #  Y: the best output label (i.e., the one with highest probability)
    #  P: probability dictionary of all classes. Its keys are class labels and its values are those labels'
    #     probabilities.
    #
    #  T: probability tensor produced by ONNX classifier
    #  T': normalized version of "T." Its sum must be 1.
    #
    # CoreML computational graph (binary class and multi-class classifications):
    #
    #            X ---> CoreML GLMClassifier ---> Y (must present in model)
    #                           |
    #                           '---------------> P
    #
    # ONNX computational graph (binary class classification):
    #
    #            X ---> ONNX GLMClassifier ---> Y (must present in model)
    #                           |
    #                           '-------------> T ---> ZipMap ---> P (If P is not specified in the considered CoreML
    #                                                                 model "T" would become an isolated variable
    #                                                                 which is not connected with any other
    #                                                                 operators. That is, both of ZipMap and P are
    #                                                                 optional in this graph.)
    #
    # ONNX computational graph (multi-class classification):
    #
    #            X ---> ONNX GLMClassifier ---> Y (must present in model)
    #                           |
    #                           '-------------> T ---> L1-norm Normalizer ---> T' ---> ZipMap ---> P
    #                                                                              (things after T' are optional.
    #                                                                               If P is specified, we may have
    #                                                                               ZipMap and P. Otherwise, this
    #                                                                               ends at T' and T' is not linked
    #                                                                               with other operators.
    from coremltools.proto.GLMClassifier_pb2 import GLMClassifier
    op_type = 'LinearClassifier'
    attrs = {'name': operator.full_name}
    zipmap_attrs = {'name': scope.get_unique_operator_name('ZipMap')}
    glm = operator.raw_operator.glmClassifier

    transform_table = {GLMClassifier.Logit: 'LOGISTIC', GLMClassifier.Probit: 'PROBIT'}
    if glm.postEvaluationTransform not in transform_table:
        raise ValueError('Unsupported post-transformation: {}'.format(glm.postEvaluationTransform))
    attrs['post_transform'] = transform_table[glm.postEvaluationTransform]

    encoding_table = {GLMClassifier.ReferenceClass: True, GLMClassifier.OneVsRest: False}
    if glm.classEncoding not in encoding_table:
        raise ValueError('Unsupported class encoding: {}'.format(glm.classEncoding))
    attrs['multi_class'] = encoding_table[glm.classEncoding]

    # Determine the dimensionality of the model weights.
    dim_target = len(glm.weights)
    dim_feature = len(glm.weights[0].value)

    matrix_w = np.ndarray(shape=(dim_target, dim_feature))
    for i, w in enumerate(glm.weights):
        matrix_w[i, :] = w.value

    if glm.WhichOneof('ClassLabels') == 'stringClassLabels':
        class_labels = list(s.encode('utf-8') for s in glm.stringClassLabels.vector)
        attrs['classlabels_strings'] = class_labels
        zipmap_attrs['classlabels_strings'] = class_labels
    elif glm.WhichOneof('ClassLabels') == 'int64ClassLabels':
        class_labels = list(int(i) for i in glm.int64ClassLabels.vector)
        attrs['classlabels_ints'] = class_labels
        zipmap_attrs['classlabels_int64s'] = class_labels
    else:
        raise ValueError('Unknown class label type')

    coefficients = matrix_w.flatten().tolist()
    intercepts = list(float(i) for i in glm.offset)
    if len(class_labels) == 2:
        # Handle the binary case for coefficients and intercepts
        coefficients = list(map(lambda x: -1 * x, coefficients)) + coefficients
        intercepts = list(map(lambda x: -1 * x, intercepts)) + intercepts

    attrs['coefficients'] = coefficients
    attrs['intercepts'] = intercepts

    # For classifiers, due to the different representation of classes' probabilities, we need to add some
    # operators for type conversion. It turns out that we have the following topology.
    # input features ---> GLMClassifier ---> label (must present)
    #                           |
    #                           '--> probability tensor ---> Normalizer ---> normalized ---> ZipMap ---> probability map
    #                                                (depending on whether probability output exists in CoreML model,
    #                                                 variables/operators after probability tensor may disappear)
    raw_model = operator.raw_operator
    # Find label name and probability name
    proba_output_name = None
    for variable in operator.outputs:
        if raw_model.description.predictedFeatureName == variable.raw_name:
            label_output_name = variable.full_name
        if raw_model.description.predictedProbabilitiesName != '' and \
                raw_model.description.predictedProbabilitiesName == variable.raw_name:
            proba_output_name = variable.full_name

    inputs = [variable.full_name for variable in operator.inputs]
    proba_tensor_name = scope.get_unique_variable_name('ProbabilityTensor')

    if proba_output_name is not None:
        # Add tree model ONNX node with probability output
        container.add_node(op_type, inputs, [label_output_name, proba_tensor_name], op_domain='ai.onnx.ml', **attrs)

        # Add a normalizer to make sure that the sum of all classes' probabilities is 1. It doesn't affect binary
        # classification. For multi-class clssifiers, if one applies sigmoid function independently to all raw scores,
        # we have to add a normalization so that the sum of all probabilities remains 1. Of course, if softmax is used
        # to convert raw scores into probabilities, this normalization doesn't change anything.
        if len(class_labels) > 2:
            normalized_proba_tensor_name = scope.get_unique_variable_name(proba_tensor_name + '_normalized')
            container.add_node('Normalizer', proba_tensor_name, normalized_proba_tensor_name, op_domain='ai.onnx.ml',
                               name=scope.get_unique_operator_name('Normalizer'), norm='L1')
        else:
            # If we don't need a normalization, we just pass the original probability tensor to the following ZipMap
            normalized_proba_tensor_name = proba_tensor_name

        # Add ZipMap to convert normalized probability tensor into probability map
        container.add_node('ZipMap', [normalized_proba_tensor_name], [proba_output_name],
                           op_domain='ai.onnx.ml', **zipmap_attrs)
    else:
        # Add linear classifier with isolated probability output, which means that the probability
        # tensor won't be accessed by any others.
        container.add_node(op_type, inputs, [label_output_name, proba_tensor_name], op_domain='ai.onnx.ml', **attrs)


register_converter('glmClassifier', convert_glm_classifier)



# SPDX-License-Identifier: Apache-2.0

import unittest
from distutils.version import StrictVersion
import numpy as np
from onnx import TensorProto
from onnx.defs import onnx_opset_version
from onnx.helper import make_node, make_graph, make_model, make_tensor_value_info, make_opsetid
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxruntime import InferenceSession, __version__ as ort_version
from onnxmltools.convert.sparkml.operator_converters.tree_helper import Node

TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmDecisionTreeClassifierBig(unittest.TestCase):

    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split(self):

        attrs = {
            'class_ids': [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
            'class_nodeids': [3, 3, 3, 4, 4, 4, 6, 6, 6, 7, 7, 7, 10, 10, 10, 11, 11, 11, 12, 12, 12],
            'class_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'class_weights': [0.4, 0.6, 0.0, 0.95, 0.04, 0.0, 0.0, 0.185, 0.814, 0.372, 0.62, 0.0, 0.74, 0.21,
                              0.03, 0.0, 1.0, 0.0, 0.87, 0.05, 0.071],
            'classlabels_int64s': [0, 1, 2],
            'name': 'TreeEnsembleClassifier',
            'nodes_falsenodeids': [8, 5, 4, 0, 0, 7, 0, 0, 12, 11, 0, 0, 0],
            'nodes_featureids': [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False, False, False, False, False, False, False, False,
                                                False, False, False, False, False],
            'nodes_modes': ['BRANCH_LEQ', '||', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'BRANCH_LEQ', 'LEAF',
                            'LEAF', 'BRANCH_LEQ', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'LEAF'],
            'nodes_nodeids': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            'nodes_values': [10.5, [55, 59, 65], 1.5, 0.0, 0.0, 8.5, 0.0, 0.0, 10.5, 9.5, 0.0, 0.0, 0.0],
            'post_transform': None}
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                  classlabels_int64s= [0, 1, 2], domain="ai.onnx.ml")
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{attrs['nodes_modes']}\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        node = make_node(op_type="TreeEnsembleClassifier", inputs=["X"], outputs=["L", "Y"], **new_attrs)
        X = make_tensor_value_info('X', TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info('Y', TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info('L', TensorProto.INT64, [None])        
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(model.SerializeToString())
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {'X': x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [2, 2, 2, 2, 2, 1, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                      classlabels_int64s= [0, 1, 2], domain="ai.onnx.ml")
        self.assertEqual(new_attrs, new_new_attrs)

    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split_non_contiguous_ids(self):

        attrs = {
            'class_ids': [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
            'class_nodeids': [3, 3, 3, 4, 4, 4, 6, 6, 6, 7, 7, 7, 10, 10, 10, 11, 11, 11, 13, 13, 13],
            'class_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'class_weights': [0.4, 0.6, 0.0, 0.95, 0.04, 0.0, 0.0, 0.185, 0.814, 0.372, 0.62, 0.0, 0.74, 0.21,
                              0.03, 0.0, 1.0, 0.0, 0.87, 0.05, 0.071],
            'classlabels_int64s': [0, 1, 2],
            'name': 'TreeEnsembleClassifier',
            'nodes_falsenodeids': [8, 5, 4, 0, 0, 7, 0, 0, 13, 11, 0, 0, 0],
            'nodes_featureids': [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False, False, False, False, False, False, False, False,
                                                False, False, False, False, False],
            'nodes_modes': ['BRANCH_LEQ', '||', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'BRANCH_LEQ', 'LEAF',
                            'LEAF', 'BRANCH_LEQ', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'LEAF'],
            'nodes_nodeids': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            'nodes_values': [10.5, [55, 59, 65], 1.5, 0.0, 0.0, 8.5, 0.0, 0.0, 10.5, 9.5, 0.0, 0.0, 0.0],
            'post_transform': None}
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                  classlabels_int64s= [0, 1, 2], domain="ai.onnx.ml")
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{attrs['nodes_modes']}\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        node = make_node(op_type="TreeEnsembleClassifier", inputs=["X"], outputs=["L", "Y"], **new_attrs)
        X = make_tensor_value_info('X', TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info('Y', TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info('L', TensorProto.INT64, [None])        
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(model.SerializeToString())
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {'X': x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [2, 2, 2, 2, 2, 1, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                      classlabels_int64s= [0, 1, 2], domain="ai.onnx.ml")
        self.assertEqual(new_attrs, new_new_attrs)

    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split_more_complex(self):

        attrs = {
            'class_ids': [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
            'class_nodeids': [3, 3, 3, 4, 4, 4, 6, 6, 6, 7, 7, 7, 10, 10, 10, 11, 11, 11, 12, 12, 12],
            'class_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'class_weights': [0.4, 0.6, 0.0, 0.95, 0.04, 0.0, 0.0, 0.185, 0.814, 0.372, 0.62, 0.0, 0.74, 0.21,
                              0.03, 0.0, 1.0, 0.0, 0.87, 0.05, 0.071],
            'classlabels_int64s': [0, 1, 2],
            'name': 'TreeEnsembleClassifier',
            'nodes_falsenodeids': [8, 5, 4, 0, 0, 7, 0, 0, 12, 11, 0, 0, 0],
            'nodes_featureids': [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            'nodes_hitrates': [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            'nodes_missing_value_tracks_true': [False, False, False, False, False, False, False, False,
                                                False, False, False, False, False],
            'nodes_modes': ['BRANCH_LEQ', 'BRANCH_LEQ', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'BRANCH_LEQ', 'LEAF',
                            'LEAF', 'BRANCH_LEQ', 'BRANCH_LEQ', 'LEAF', 'LEAF', 'LEAF'],
            'nodes_nodeids': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            'nodes_treeids': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'nodes_truenodeids': [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            'nodes_values': [[100, 101, 102], [55, 59, 65], [10, 11, 12], 0.0, 0.0, [1000, 1001, 1002],
                             0.0, 0.0, [10000, 10001], [100000, 100001], 0.0, 0.0, 0.0],
            'post_transform': None}
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                  classlabels_int64s=[0, 1, 2], domain="ai.onnx.ml")
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{attrs['nodes_modes']}\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        self.assertEqual(len(new_attrs["nodes_modes"]), len(attrs["nodes_modes"]) + 10)
        node = make_node(op_type="TreeEnsembleClassifier", inputs=["X"], outputs=["L", "Y"], **new_attrs)
        X = make_tensor_value_info('X', TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info('Y', TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info('L', TensorProto.INT64, [None])        
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(model.SerializeToString())
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {'X': x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                      classlabels_int64s= [0, 1, 2], domain="ai.onnx.ml")
        self.assertEqual(new_attrs, new_new_attrs)

    def test_debug(self):
        try:
            from onnxmltools.debug1 import attrs
        except ImportError as e:
            return
        root, nodes = Node.create(attrs)
        root.unfold_rule_or()
        new_attrs = root.to_attrs(post_transform=None, name="TreeEnsembleClassifier",
                                  classlabels_int64s=[0, 1, 2], domain="ai.onnx.ml")
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{len(new_attrs[k])}\nbefore {len(attrs[k])}."
                )


if __name__ == "__main__":
    # TestSparkmDecisionTreeClassifierBig().test_debug()
    unittest.main(verbosity=2)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_mean
from ....common._registration import register_converter


def convert_average(scope, operator, container):
    apply_mean(scope, operator.input_full_names, operator.output_full_names, container,
               operator_name=operator.full_name)


register_converter('average', convert_average)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_batch_normalization_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C] ---> [N, C]
        2. [N, C, H, W] ---> [N, C, H, W]

    This operator just uses the operator input shape as its output shape.
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input_shape = operator.inputs[0].type.shape
    if len(input_shape) not in [2, 4]:
        raise RuntimeError('Input must be a 2-D or a 4-D tensor')

    operator.outputs[0].type.shape = copy.deepcopy(operator.inputs[0].type.shape)


register_shape_calculator('batchnorm', calculate_batch_normalization_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter

COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE = {
    0: 'BRANCH_LEQ',
    1: 'BRANCH_LT',
    2: 'BRANCH_GTE',
    3: 'BRANCH_GT',
    4: 'BRANCH_EQ',
    5: 'BRANCH_NEQ',
    6: 'LEAF'
}

COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM = {
    0: 'NONE',
    1: 'SOFTMAX',
    2: 'LOGISTIC',
    3: 'SOFTMAX_ZERO'
}


def get_onnx_tree_mode(cm_tree_behavior):
    if cm_tree_behavior in COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE:
        return COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE[cm_tree_behavior]
    raise ValueError('CoreML tree node behavior not supported {0}'.format(cm_tree_behavior))


def get_onnx_tree_post_transform(cm_tree_post_transform):
    if cm_tree_post_transform in COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM:
        return COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM[cm_tree_post_transform]
    raise ValueError('CoreML tree post transform not supported {0}'.format(cm_tree_post_transform))


def convert_tree_ensemble_model(scope, operator, container):
    raw_model = operator.raw_operator
    attrs = {'name': operator.full_name}
    if raw_model.WhichOneof('Type') == 'treeEnsembleClassifier':
        op_type = 'TreeEnsembleClassifier'
        prefix = 'class'
        nodes = raw_model.treeEnsembleClassifier.treeEnsemble.nodes
        attrs['base_values'] = raw_model.treeEnsembleClassifier.treeEnsemble.basePredictionValue
        attrs['post_transform'] = get_onnx_tree_post_transform(raw_model.treeEnsembleClassifier.postEvaluationTransform)
        zipmap_attrs = {'name': scope.get_unique_operator_name('ZipMap')}
        if raw_model.treeEnsembleClassifier.WhichOneof('ClassLabels') == 'int64ClassLabels':
            class_labels = list(int(i) for i in raw_model.treeEnsembleClassifier.int64ClassLabels.vector)
            attrs['classlabels_int64s'] = class_labels
            zipmap_attrs['classlabels_int64s'] = class_labels
        else:
            class_labels = list(s.encode('utf-8') for s in raw_model.treeEnsembleClassifier.stringClassLabels.vector)
            attrs['classlabels_strings'] = class_labels
            zipmap_attrs['classlabels_strings'] = class_labels
    elif raw_model.WhichOneof('Type') == 'treeEnsembleRegressor':
        op_type = 'TreeEnsembleRegressor'
        prefix = 'target'
        nodes = raw_model.treeEnsembleRegressor.treeEnsemble.nodes
        attrs['base_values'] = raw_model.treeEnsembleRegressor.treeEnsemble.basePredictionValue
        attrs['n_targets'] = raw_model.treeEnsembleRegressor.treeEnsemble.numPredictionDimensions
        attrs['post_transform'] = get_onnx_tree_post_transform(raw_model.treeEnsembleRegressor.postEvaluationTransform)
    else:
        raise ValueError('Unknown tree model type')

    leaf_treeids = [node.treeId for node in nodes if 6 == node.nodeBehavior for _ in node.evaluationInfo]
    leaf_nodeids = [node.nodeId for node in nodes if 6 == node.nodeBehavior for _ in node.evaluationInfo]
    leaf_ids = [weight.evaluationIndex for node in nodes if 6 == node.nodeBehavior for weight in node.evaluationInfo]

    leaf_weights = [weight.evaluationValue for node in nodes if 6 == node.nodeBehavior for weight in
                    node.evaluationInfo]

    assert (len(leaf_ids) == len(leaf_weights))
    assert (len(leaf_weights) == len(leaf_nodeids))
    assert (len(leaf_nodeids) == len(leaf_treeids))

    nodes_nodeids = [x.nodeId for x in nodes]
    nodes_treeids = [x.treeId for x in nodes]
    nodes_featureids = [x.branchFeatureIndex for x in nodes]
    nodes_values = [x.branchFeatureValue for x in nodes]
    nodes_truenodeids = [x.trueChildNodeId for x in nodes]
    nodes_falsenodeids = [x.falseChildNodeId for x in nodes]
    nodes_missing_value_tracks_true = [x.missingValueTracksTrueChild for x in nodes]
    nodes_hitrates = [float(x.relativeHitRate) for x in nodes]
    nodes_modes = [get_onnx_tree_mode(x.nodeBehavior) for x in nodes]

    attrs['nodes_treeids'] = nodes_treeids
    attrs['nodes_nodeids'] = nodes_nodeids
    attrs['nodes_featureids'] = nodes_featureids
    attrs['nodes_values'] = nodes_values
    attrs['nodes_hitrates'] = nodes_hitrates
    attrs['nodes_modes'] = nodes_modes
    attrs['nodes_truenodeids'] = nodes_truenodeids
    attrs['nodes_falsenodeids'] = nodes_falsenodeids
    attrs['nodes_missing_value_tracks_true'] = nodes_missing_value_tracks_true
    attrs[prefix + '_treeids'] = leaf_treeids
    attrs[prefix + '_nodeids'] = leaf_nodeids
    attrs[prefix + '_ids'] = leaf_ids
    attrs[prefix + '_weights'] = leaf_weights

    # For regression, we can simply construct a model. For classifier, due to the different representation of
    # classes' probabilities, we need to add some operators for type conversion.
    if raw_model.WhichOneof('Type') == 'treeEnsembleRegressor':
        # Create ONNX representation of this operator. If there is only one input, its full topology is
        #
        # input features ---> TreeEnsembleRegressor ---> output
        #
        # If there are multiple (e.g., "N" features) input features, we need to concatenate them all together before feeding them into
        # ONNX tree-based model. It leads to the following computational graph.
        #
        # input feature 1 -----.
        #        ...           |
        #        ...           v
        #        ...      ---> Feature Vectorizer ---> TreeEnsembleRegressor ---> output
        #        ...           ^
        #        ...           |
        # input feature N -----'
        if len(operator.inputs) > 1:
            feature_vector_name = scope.get_unique_variable_name('feature_vector')
            container.add_node('FeatureVectorizer', operator.input_full_names, feature_vector_name,
                               op_domain='ai.onnx.ml', name=scope.get_unique_operator_name('FeatureVectorizer'),
                               inputdimensions=[variable.type.shape[1] for variable in operator.inputs])
            container.add_node(op_type, feature_vector_name, operator.output_full_names,
                               op_domain='ai.onnx.ml', **attrs)
        else:
            container.add_node(op_type, operator.input_full_names, operator.output_full_names,
                               op_domain='ai.onnx.ml', **attrs)
    else:
        # For classifiers, due to the different representation of classes' probabilities, we need to add some
        # operators for type conversion. It turns out that we have the following topology.
        # input features ---> TreeEnsembleClassifier ---> label (must present)
        #                               |
        #                               '--> probability tensor ---> ZipMap ---> probability map (optional)
        #
        # Similar to the regressor's case, if there are multiple input features, we need to concatenate them all
        # together before feeding them into ONNX tree-based model. It leads to the following computational graph.
        #
        # input feature 1 -----.
        #        ...           |
        #        ...           v
        #        ...      ---> Feature Vectorizer ---> TreeEnsembleClassifier ---> label (must present)
        #        ...           ^                                 |
        #        ...           |                                 '--> probability tensor ---> ZipMap ---> probability
        # input feature N -----'                                                                          map (optional)

        # Set up input feature(s)
        if len(operator.inputs) > 1:
            feature_vector_name = scope.get_unique_variable_name('feature_vector')
            container.add_node('FeatureVectorizer', operator.input_full_names, feature_vector_name,
                               op_domain='ai.onnx.ml', name=scope.get_unique_operator_name('FeatureVectorizer'),
                               inputdimensions=[variable.type.shape[1] for variable in operator.inputs])
        else:
            feature_vector_name = operator.inputs[0].full_name

        # Find label name and probability name
        proba_output_name = None
        for variable in operator.outputs:
            if raw_model.description.predictedFeatureName == variable.raw_name:
                label_output_name = variable.full_name
            if raw_model.description.predictedProbabilitiesName != '' and raw_model.description.predictedProbabilitiesName == variable.raw_name:
                proba_output_name = variable.full_name

        proba_tensor_name = scope.get_unique_variable_name('ProbabilityTensor')

        if proba_output_name is not None:
            # Add tree model ONNX node with probability output
            container.add_node(op_type, feature_vector_name, [label_output_name, proba_tensor_name],
                               op_domain='ai.onnx.ml', **attrs)

            # Add ZipMap to convert probability tensor into probability map
            container.add_node('ZipMap', [proba_tensor_name], [proba_output_name],
                               op_domain='ai.onnx.ml', **zipmap_attrs)
        else:
            # Add support vector classifier without probability output
            container.add_node(op_type, feature_vector_name, [label_output_name, proba_tensor_name],
                               op_domain='ai.onnx.ml', **attrs)


register_converter("treeEnsembleClassifier", convert_tree_ensemble_model)
register_converter("treeEnsembleRegressor", convert_tree_ensemble_model)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import DictionaryType, FloatTensorType, SequenceType, StringTensorType, Int64TensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_tensor_to_probability_map_output_shapes(operator):
    '''
    Allowed input/output patterns are
    ONNX < 1.2
        1. [1, C] ---> ---> A map
        2. [1, C_1, ..., C_n] ---> A map
    ONNX >= 1.2
        1. [N, C] ---> ---> A sequence of maps
        2. [N, C_1, ..., C_n] ---> A sequence of maps

    Note that N must be 1 currently if you're using ONNX<1.2 because old ZipMap doesn't produce a seqneuce of map If the
    input is not [N, C], it will be reshaped into [N, C_1 x C_2, x ... x C_n] before being fed into ONNX ZipMap.
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    model_type = operator.raw_operator.WhichOneof('Type')
    if model_type == 'neuralNetworkClassifier':
        class_label_type = operator.raw_operator.neuralNetworkClassifier.WhichOneof('ClassLabels')
    else:
        raise TypeError('%s has no class label' % model_type)

    N = operator.inputs[0].type.shape[0]
    doc_string = operator.outputs[0].type.doc_string
    if class_label_type == 'stringClassLabels':
        if operator.target_opset < 7:
            operator.outputs[0].type = DictionaryType(StringTensorType([1]), FloatTensorType([1]), doc_string)
        else:
            operator.outputs[0].type = \
                SequenceType(DictionaryType(StringTensorType([]), FloatTensorType([])), N, doc_string)
    elif class_label_type == 'int64ClassLabels':
        if operator.target_opset < 7:
            operator.outputs[0].type = DictionaryType(Int64TensorType([1]), FloatTensorType([1]), doc_string)
        else:
            operator.outputs[0].type = \
                SequenceType(DictionaryType(Int64TensorType([]), FloatTensorType([])), N, doc_string)
    else:
        raise ValueError('Unsupported label type')


register_shape_calculator('tensorToProbabilityMap', calculate_tensor_to_probability_map_output_shapes)



# SPDX-License-Identifier: Apache-2.0

try:
    from onnxconverter_common.utils import hummingbird_installed  # noqa
except ImportError:
    def hummingbird_installed():
        """
        Checks that *Hummingbird* is available.
        """
        try:
            import hummingbird.ml  # noqa: F401

            return True
        except ImportError:
            return False


def tf2onnx_installed():
    """
    Checks that *tf2onnx* is available.
    """
    try:
        import tf2onnx  # noqa F401
        return True
    except ImportError:
        return False


from onnxconverter_common.utils import *  # noqa



# SPDX-License-Identifier: Apache-2.0

import numpy
from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_floor, apply_div, apply_matmul
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from .tree_ensemble_common import save_read_sparkml_model_data

g_rand_vectors = None


def get_rand_vectors(operator):
    global g_rand_vectors
    if not g_rand_vectors:
        g_rand_vectors = save_read_sparkml_model_data(
            operator.raw_params['SparkSession'], operator.raw_operator
        ).first()[0].toArray().transpose().astype(numpy.float32)
    return g_rand_vectors


def convert_min_hash_lsh(scope, operator, container):
    rand_vectors = get_rand_vectors(operator)
    bucket_length = float(operator.raw_operator.getOrDefault('bucketLength'))

    rand_vectors_tensor = scope.get_unique_variable_name('rand_vectors_tensor')
    container.add_initializer(rand_vectors_tensor, onnx_proto.TensorProto.FLOAT,
                              rand_vectors.shape, rand_vectors.flatten())
    matmul_result = scope.get_unique_variable_name('matmul_result_tensor')
    apply_matmul(scope, [operator.input_full_names[0], rand_vectors_tensor], matmul_result, container)
    bucket_length_tensor = scope.get_unique_variable_name('bucket_length_tensor')
    container.add_initializer(bucket_length_tensor, onnx_proto.TensorProto.FLOAT,
                              [1], [bucket_length])
    div_result = scope.get_unique_variable_name('div_result_tensor')
    apply_div(scope, [matmul_result, bucket_length_tensor], div_result, container)
    apply_floor(scope, div_result, operator.output_full_names[0], container)


register_converter('pyspark.ml.feature.BucketedRandomProjectionLSHModel', convert_min_hash_lsh)


def calculate_min_hash_lsh_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    N = operator.inputs[0].type.shape[0]
    C = 1
    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator('pyspark.ml.feature.BucketedRandomProjectionLSHModel', calculate_min_hash_lsh_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
from pyspark.ml.feature import StringIndexer
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase
import numpy

TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlStringIndexer(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_string_indexer(self):
        indexer = StringIndexer(inputCol="cat1", outputCol="cat1_index", handleInvalid="skip")
        data = self.spark.createDataFrame([("a",), ("b",), ("c",), ("a",), ("a",), ("c",)], ["cat1"])
        model = indexer.fit(data)
        # the input name should match that of what StringIndexer.inputCol
        model_onnx = convert_sparkml(
            model, "Sparkml StringIndexer", [("cat1", StringTensorType([None, 1]))], target_opset=TARGET_OPSET
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(data)
        expected = predicted.select("cat1_index").toPandas().values
        data_np = data.select("cat1").toPandas().values
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlStringIndexer")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["cat1_index"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_string_indexer_multiple_cols(self):
        indexer = StringIndexer(
            inputCols=["cat1", "cat2"], outputCols=["cat1_index", "cat2_index"], handleInvalid="skip"
        )
        data = self.spark.createDataFrame(
            [("a", "x",), ("b", "x",), ("c", "y",), ("a", "y",), ("a", "z",), ("c", "z",)], ["cat1", "cat2"]
        )
        model = indexer.fit(data)
        # the input name should match that of what StringIndexer.inputCol
        model_onnx = convert_sparkml(
            model,
            "Sparkml StringIndexer",
            [("cat1", StringTensorType([None, 1])), ("cat2", StringTensorType([None, 1]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(data)

        expected = {
            "cat1_index": predicted.select("cat1_index").toPandas().values.astype(numpy.int64),
            "cat2_index": predicted.select("cat2_index").toPandas().values.astype(numpy.int64),
        }
        data_np = {
            "cat1": data.select("cat1").toPandas().values.astype(str),
            "cat2": data.select("cat2").toPandas().values.astype(str),
        }
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlStringIndexer")
        onnx_model_path = paths[-1]
        output_names = ["cat1_index", "cat2_index"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))
        compare_results(expected, actual_output, decimal=5)
        assert output_shapes == [[None, 1], [None, 1]]


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common._registration import register_converter, register_shape_calculator
from ...common.utils import check_input_and_output_numbers
from ...common.data_types import *


def convert_sparkml_vector_indexer(scope, operator, container):
    feature_count = operator.raw_operator.numFeatures
    category_map = operator.raw_operator.categoryMaps
    split_output_names = [ scope.get_unique_variable_name('split_tensor_%d' % i) for i in range(0, feature_count)]
    if feature_count > 1:
        container.add_node('Split', operator.inputs[0].full_name, split_output_names,
                           name=scope.get_unique_operator_name('Split'),
                           op_version=2,
                           axis=1,
                           split=[1]*feature_count)
    else:
        split_output_names = operator.input_full_names
    concat_inputs = split_output_names.copy()
    for i in category_map.keys():
        converted_output = scope.get_unique_variable_name('converted_tensor_%d' % i)
        container.add_node('Cast', split_output_names[i], converted_output,
                           name=scope.get_unique_operator_name('Cast'),
                           op_version=9,
                           to=onnx_proto.TensorProto.STRING)
        attrs = {
            'name': scope.get_unique_operator_name('LabelEncoder'),
            'classes_strings': ['{0:g}'.format(c) for c in category_map[i].keys()],
            'default_string': '__unknown__'
        }
        encoded_output_name = scope.get_unique_variable_name('indexed_tensor_%d' % i)
        container.add_node('LabelEncoder', converted_output, encoded_output_name,
                       op_domain='ai.onnx.ml',
                       **attrs)
        converted_float_output = scope.get_unique_variable_name('converted_float_tensor_%d' % i)
        if feature_count == 1:
            converted_float_output = operator.output_full_names[0]
        container.add_node('Cast', encoded_output_name, converted_float_output,
                           name=scope.get_unique_operator_name('Cast'),
                           op_version=9,
                           to=onnx_proto.TensorProto.FLOAT)
        concat_inputs[i] = converted_float_output
    # add the final Concat
    if feature_count > 1:
        container.add_node('Concat', concat_inputs, operator.output_full_names[0],
                           name=scope.get_unique_operator_name('Concat'),
                           op_version=4,
                           axis=1)


register_converter('pyspark.ml.feature.VectorIndexerModel', convert_sparkml_vector_indexer)


def calculate_vector_indexer_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, operator.raw_operator.numFeatures])


register_shape_calculator('pyspark.ml.feature.VectorIndexerModel', calculate_vector_indexer_shapes)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from pyspark.ml.classification import GBTClassificationModel

from ...common._apply_operation import apply_neg, apply_concat, apply_mul, apply_exp, apply_add, \
    apply_argmax, apply_matmul
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator
from ..ops_names import get_sparkml_operator_name
from .decision_tree_regressor import convert_decision_tree_regressor


def convert_gbt_classifier(scope, operator, container):
    op = operator.raw_operator
    regressor_output_names = []
    # spark implementation uses DecisionTreeRegressor (and not Classifier) for each tree in this forest
    for tree_model in op.trees:
        regressor_op = scope.declare_local_operator(get_sparkml_operator_name(type(tree_model)), tree_model)
        regressor_op.raw_params = operator.raw_params
        regressor_op.inputs = operator.inputs
        regressor_output = scope.declare_local_variable('regressor_prediction', FloatTensorType())
        regressor_output_names.append(regressor_output.full_name)
        regressor_op.outputs.append(regressor_output)
        convert_decision_tree_regressor(scope, regressor_op, container)
        regressor_op.is_evaluated = True

    targets_tensor = scope.get_unique_variable_name('target_tensor')
    weights_tensor = scope.get_unique_variable_name('weights_tensor')
    container.add_initializer(weights_tensor, onnx_proto.TensorProto.FLOAT, [len(op.treeWeights), 1], op.treeWeights)
    concatenated_predictions = scope.get_unique_variable_name('concatenated_predictions_tensor')
    apply_concat(scope, regressor_output_names, concatenated_predictions, container, axis=1)
    apply_matmul(scope, [concatenated_predictions, weights_tensor], targets_tensor, container)

    # this is to calculate prediction and probability given the raw_prediction (= [-target, target])
    targets_neg_tensor = scope.get_unique_variable_name('target_neg_tensor')
    apply_neg(scope, targets_tensor, targets_neg_tensor, container)
    raw_prediction_tensor = scope.get_unique_variable_name('raw_prediction_tensor')
    apply_concat(scope, [targets_neg_tensor, targets_tensor], raw_prediction_tensor, container,
                 axis=1)
    if isinstance(op, GBTClassificationModel):
        # this section is only for the classifier; for the regressor we don't calculate the probability
        minus_two = scope.get_unique_variable_name('minus_two_tensor')
        container.add_initializer(minus_two, onnx_proto.TensorProto.FLOAT, [1], [-2.0])
        mul_output_tensor = scope.get_unique_variable_name('mul_output_tensor')
        apply_mul(scope, [raw_prediction_tensor, minus_two], mul_output_tensor, container)
        exp_output_tensor = scope.get_unique_variable_name('exp_output_tensor')
        apply_exp(scope, mul_output_tensor, exp_output_tensor, container)
        one_tensor = scope.get_unique_variable_name('one_tensor')
        container.add_initializer(one_tensor, onnx_proto.TensorProto.FLOAT, [1], [1.0])
        add_output_tensor = scope.get_unique_variable_name('add_output_tensor')
        apply_add(scope, [exp_output_tensor, one_tensor], add_output_tensor, container)
        container.add_node('Reciprocal', add_output_tensor, operator.outputs[1].full_name,
                           name=scope.get_unique_operator_name('Reciprocal'),
                           op_version=6)
    # to get Prediction from rawPrediction (or probability)
    apply_argmax(scope, raw_prediction_tensor, operator.outputs[0].full_name, container,
                 axis=1, keepdims=0)


register_converter('pyspark.ml.classification.GBTClassificationModel', convert_gbt_classifier)
register_converter('pyspark.ml.regression.GBTRegressionModel', convert_gbt_classifier)


def calculate_gbt_classifier_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=[1, 2])
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType])
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Input must be a [N, C]-tensor')

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])
    if isinstance(operator.raw_operator, GBTClassificationModel):
        class_count = 2
        operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator('pyspark.ml.classification.GBTClassificationModel',
                          calculate_gbt_classifier_output_shapes)
register_shape_calculator('pyspark.ml.regression.GBTRegressionModel',
                          calculate_gbt_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import tempfile
import os
import time
import numpy
from pyspark.sql import SparkSession

class SparkMLTree(dict):
    pass


def sparkml_tree_dataset_to_sklearn(tree_df, is_classifier):
    feature = []
    threshold = []
    tree_pandas = tree_df.toPandas().sort_values("id")
    children_left = tree_pandas.leftChild.values.tolist()
    children_right = tree_pandas.rightChild.values.tolist()
    ids = tree_pandas.id.values.tolist()
    if is_classifier:
        value = numpy.array(tree_pandas.impurityStats.values.tolist())
    else:
        value = tree_pandas.prediction.values.tolist()

    for item in tree_pandas.split:
        if isinstance(item, dict):
            try:
                feature.append(item["featureIndex"])
                threshold.append(item["leftCategoriesOrThreshold"])
            except KeyError as e:
                raise RuntimeError(f"Unable to process {item}.")
        else:
            tuple_item = tuple(item)
            feature.append(item[0])
            threshold.append(item[1][0] if len(item[1]) >= 1 else -1.0)

    tree = SparkMLTree()
    tree.nodes_ids = ids
    tree.children_left = children_left
    tree.children_right = children_right
    tree.value = numpy.asarray(value, dtype=numpy.float32)
    tree.feature = feature
    tree.threshold = threshold
    tree.node_count = tree_df.count()
    return tree


def save_read_sparkml_model_data(spark: SparkSession, model):
    tdir = tempfile.tempdir
    if tdir is None:
        local_dir = spark._jvm.org.apache.spark.util.Utils.getLocalDir(spark._jsc.sc().conf())
        tdir = spark._jvm.org.apache.spark.util.Utils.createTempDir(local_dir, "onnx").getAbsolutePath()
    if tdir is None:
        raise FileNotFoundError(
            "Unable to create a temporary directory for model '{}'"
            ".".format(type(model).__name__))
    path = os.path.join(tdir, type(model).__name__ + "_" + str(time.time()))
    model.write().overwrite().save(path)
    df = spark.read.parquet(os.path.join(path, 'data'))
    return df


def get_default_tree_classifier_attribute_pairs():
    attrs = {}
    attrs['post_transform'] = 'NONE'
    attrs['nodes_treeids'] = []
    attrs['nodes_nodeids'] = []
    attrs['nodes_featureids'] = []
    attrs['nodes_modes'] = []
    attrs['nodes_values'] = []
    attrs['nodes_truenodeids'] = []
    attrs['nodes_falsenodeids'] = []
    attrs['nodes_missing_value_tracks_true'] = []
    attrs['nodes_hitrates'] = []
    attrs['class_treeids'] = []
    attrs['class_nodeids'] = []
    attrs['class_ids'] = []
    attrs['class_weights'] = []
    return attrs


def get_default_tree_regressor_attribute_pairs():
    attrs = {}
    attrs['post_transform'] = 'NONE'
    attrs['n_targets'] = 0
    attrs['nodes_treeids'] = []
    attrs['nodes_nodeids'] = []
    attrs['nodes_featureids'] = []
    attrs['nodes_modes'] = []
    attrs['nodes_values'] = []
    attrs['nodes_truenodeids'] = []
    attrs['nodes_falsenodeids'] = []
    attrs['nodes_missing_value_tracks_true'] = []
    attrs['nodes_hitrates'] = []
    attrs['target_treeids'] = []
    attrs['target_nodeids'] = []
    attrs['target_ids'] = []
    attrs['target_weights'] = []
    return attrs


def add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id, feature_id, mode, value, true_child_id,
             false_child_id, weights, weight_id_bias, leaf_weights_are_counts):
    attr_pairs['nodes_treeids'].append(tree_id)
    attr_pairs['nodes_nodeids'].append(node_id)
    attr_pairs['nodes_featureids'].append(feature_id)
    attr_pairs['nodes_modes'].append(mode)
    attr_pairs['nodes_values'].append(value)
    attr_pairs['nodes_truenodeids'].append(true_child_id)
    attr_pairs['nodes_falsenodeids'].append(false_child_id)
    attr_pairs['nodes_missing_value_tracks_true'].append(False)
    attr_pairs['nodes_hitrates'].append(1.)

    # Add leaf information for making prediction
    if mode == 'LEAF':
        flattened_weights = weights.flatten()
        factor = tree_weight
        # If the values stored at leaves are counts of possible classes, we need convert them to probabilities by
        # doing a normalization.
        if leaf_weights_are_counts:
            s = sum(flattened_weights)
            factor /= float(s) if s != 0. else 1.
        flattened_weights = [w * factor for w in flattened_weights]
        if len(flattened_weights) == 2 and is_classifier:
            flattened_weights = [flattened_weights[1]]

        # Note that attribute names for making prediction are different for classifiers and regressors
        if is_classifier:
            for i, w in enumerate(flattened_weights):
                attr_pairs['class_treeids'].append(tree_id)
                attr_pairs['class_nodeids'].append(node_id)
                attr_pairs['class_ids'].append(i + weight_id_bias)
                attr_pairs['class_weights'].append(w)
        else:
            for i, w in enumerate(flattened_weights):
                attr_pairs['target_treeids'].append(tree_id)
                attr_pairs['target_nodeids'].append(node_id)
                attr_pairs['target_ids'].append(i + weight_id_bias)
                attr_pairs['target_weights'].append(w)


def add_tree_to_attribute_pairs(attr_pairs, is_classifier, tree, tree_id, tree_weight,
                                weight_id_bias, leaf_weights_are_counts):
    for i in range(tree.node_count):
        node_id = tree.nodes_ids[i]
        weight = tree.value[i]

        if tree.children_left[i] >= 0 or tree.children_right[i] >= 0:
            mode = 'BRANCH_LEQ'
            feat_id = tree.feature[i]
            threshold = tree.threshold[i]
            left_child_id = int(tree.children_left[i])
            right_child_id = int(tree.children_right[i])
        else:
            mode = 'LEAF'
            feat_id = 0
            threshold = 0.
            left_child_id = 0
            right_child_id = 0

        add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id, feat_id, mode, threshold,
                 left_child_id, right_child_id, weight, weight_id_bias, leaf_weights_are_counts)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import Int64Type, Int64TensorType
from ...common._registration import register_converter


def convert_feature_vectorizer(scope, operator, container):
    op_type = 'FeatureVectorizer'
    attrs = {'name': operator.full_name}

    inputs = []
    input_dims = []
    for variable in operator.inputs:
        if type(variable.type) in [Int64TensorType, Int64Type]:
            # We use scaler to convert integers into floats because output is a single tensor and all tensor elements
            # should be in the same type.
            scaler_name = scope.get_unique_operator_name('Scaler')
            scaled_name = scope.get_unique_variable_name(variable.full_name + '_scaled')
            scaler_attrs = {'name': scaler_name, 'scale': [1.], 'offset': [0.]}
            container.add_node('Scaler', [variable.full_name], [scaled_name], op_domain='ai.onnx.ml', **scaler_attrs)
            inputs.append(scaled_name)
        else:
            inputs.append(variable.full_name)
        # We assume feature vectorizer always combines inputs with shapes [1, C] or [C]
        input_dims.append(variable.type.shape[1])
    attrs['inputdimensions'] = input_dims

    container.add_node(op_type, inputs, [operator.outputs[0].full_name], op_domain='ai.onnx.ml', **attrs)


register_converter('featureVectorizer', convert_feature_vectorizer)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_add


def convert_add(scope, operator, container):
    if len(operator.input_full_names) == 1:
        scaler_name = scope.get_unique_variable_name(operator.full_name + '_B')
        container.add_initializer(scaler_name, onnx_proto.TensorProto.FLOAT, [], [operator.raw_operator.add.alpha])
        inputs = [operator.inputs[0].full_name, scaler_name]
        broadcast = 1
        apply_add(scope, inputs, operator.output_full_names, container, operator_name=operator.full_name,
                  broadcast=broadcast)
    else:
        inputs = operator.input_full_names

        if len(inputs) == 2:
            left_tensor = operator.inputs[0].full_name
            right_tensor = operator.inputs[1].full_name
            if operator.inputs[0].type.shape != operator.inputs[1].type.shape:
                broadcast = 1
            else:
                broadcast = 0

            apply_add(scope, [left_tensor, right_tensor], operator.outputs[0].full_name, container,
                      operator_name=operator.full_name, broadcast=broadcast)
        else:
            left_tensor = operator.inputs[0].full_name
            right_tensor = operator.inputs[1].full_name

            # Sum up the first two inputs
            intermediate_tensor_name = scope.get_unique_variable_name('buffer_tensor')
            apply_add(scope, [left_tensor, right_tensor], intermediate_tensor_name, container,
                      operator_name=operator.full_name, broadcast=1)

            # Accumulate other inputs onto intermediate tensors. Note that we may use the original operator's output as
            # the last intermediate tensor.
            for i in range(2, len(inputs)):
                left_tensor = intermediate_tensor_name
                right_tensor = inputs[i].full_name
                if i != len(inputs) - 1:
                    intermediate_tensor_name = scope.get_unique_variable_name('buffer_tensor')
                else:
                    intermediate_tensor_name = operator.outputs[0].full_name
                apply_add(scope, [left_tensor, right_tensor], intermediate_tensor_name, container, broadcast=1)


register_converter('add', convert_add)



# SPDX-License-Identifier: Apache-2.0

"""
Common function to converters and shape calculators.
"""

def get_xgb_params(xgb_node):
    """
    Retrieves parameters of a model.
    """
    if hasattr(xgb_node, 'kwargs'):
        # XGBoost >= 0.7
        params = xgb_node.get_xgb_params()
    else:
        # XGBoost < 0.7
        params = xgb_node.__dict__

    if ('n_estimators' not in params and
            hasattr(xgb_node, 'n_estimators')):
        # xgboost >= 1.0.2
        params['n_estimators'] = xgb_node.n_estimators
    return params



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter

_LINK_FUNCTION_TO_POST_TRANSFORM = {
    'identity': 'NONE',
    'logit': 'LOGISTIC',
    'ologit': 'LOGISTIC'
}


def _get_post_transform(params):
    link_function = params["link_function"]
    family = params["family"]
    if family == "multinomial":
        return 'SOFTMAX'
    elif link_function not in _LINK_FUNCTION_TO_POST_TRANSFORM.keys():
        raise ValueError("Link function %s not supported." % link_function)
    else:
        return _LINK_FUNCTION_TO_POST_TRANSFORM[link_function]


def _get_default_tree_attribute_pairs(is_classifier, params):
    attrs = {
        'post_transform': _get_post_transform(params)
    }
    nclasses = params["nclasses"]
    if is_classifier:
        predicted_classes = nclasses if nclasses > 2 else 1
        attrs['base_values'] = [params["base_score"] for _ in range(0, predicted_classes)]
    else:
        attrs['n_targets'] = 1
        attrs['base_values'] = [params["base_score"]]
    for k in {'nodes_treeids',  'nodes_nodeids',
              'nodes_featureids', 'nodes_modes', 'nodes_values',
              'nodes_truenodeids', 'nodes_falsenodeids', 'nodes_missing_value_tracks_true'}:
        attrs[k] = []
    node_attr_prefix = _node_attr_prefix(is_classifier)
    for k in {'_treeids', '_nodeids', '_ids', '_weights'}:
        attrs[node_attr_prefix + k] = []
    return attrs


def _add_node(
        attr_pairs, is_classifier, tree_id, node_id,
        feature_id, mode, value, true_child_id, false_child_id, weights,
        missing
):
    attr_pairs['nodes_treeids'].append(tree_id)
    attr_pairs['nodes_nodeids'].append(node_id)
    attr_pairs['nodes_featureids'].append(feature_id)
    attr_pairs['nodes_modes'].append(mode)
    attr_pairs['nodes_values'].append(float(value))
    attr_pairs['nodes_truenodeids'].append(true_child_id)
    attr_pairs['nodes_falsenodeids'].append(false_child_id)
    attr_pairs['nodes_missing_value_tracks_true'].append(missing)
    if mode == 'LEAF':
        node_attr_prefix = _node_attr_prefix(is_classifier)
        for i, w in enumerate(weights):
            attr_pairs[node_attr_prefix + '_treeids'].append(tree_id)
            attr_pairs[node_attr_prefix + '_nodeids'].append(node_id)
            attr_pairs[node_attr_prefix + '_ids'].append(i)
            attr_pairs[node_attr_prefix + '_weights'].append(float(w))


def _node_attr_prefix(is_classifier):
    return "class" if is_classifier else "target"


def _fill_node_attributes(tree_id, node, attr_pairs, is_classifier):
    if 'leftChild' in node:
        if node["isCategorical"]:
            raise ValueError("categorical splits not supported, use one_hot_explicit")
        else:
            operator = 'BRANCH_GTE'
            value = node['splitValue']
        _add_node(
            attr_pairs=attr_pairs,
            is_classifier=is_classifier,
            tree_id=tree_id,
            mode=operator,
            value=value,
            node_id=node['id'],
            feature_id=node['colId'],
            true_child_id=node['rightChild']['id'],
            false_child_id=node['leftChild']['id'],
            weights=None,
            missing=(0 if node["leftward"] else 1),
        )
        _fill_node_attributes(tree_id, node["leftChild"], attr_pairs, is_classifier)
        _fill_node_attributes(tree_id, node["rightChild"], attr_pairs, is_classifier)
    else:  # leaf
        weights = [node['predValue']]
        _add_node(
            attr_pairs=attr_pairs,
            is_classifier=is_classifier,
            tree_id=tree_id,
            value=0.,
            node_id=node['id'],
            feature_id=0, mode='LEAF',
            true_child_id=0, false_child_id=0,
            weights=weights,
            missing=False
        )


def assign_node_ids(node, next_id):
    if node is None:
        return next_id
    node["id"] = next_id
    next_id += 1
    next_id = assign_node_ids(node.get("leftChild", None), next_id)
    return assign_node_ids(node.get("rightChild", None), next_id)


def fill_tree_attributes(model, attr_pairs, node_attr_prefix):
    for tree in model["trees"]:
        assign_node_ids(tree["root"], 0)
        _fill_node_attributes(tree["index"], tree["root"], attr_pairs, node_attr_prefix)


def convert_regression(scope, operator, container, params):
    model = operator.raw_operator

    attr_pairs = _get_default_tree_attribute_pairs(False, params)
    fill_tree_attributes(model, attr_pairs, False)

    # add nodes
    container.add_node('TreeEnsembleRegressor', operator.input_full_names,
                       operator.output_full_names, op_domain='ai.onnx.ml',
                       name=scope.get_unique_operator_name('TreeEnsembleRegressor'), **attr_pairs)


def convert_classifier(scope, operator, container, params):
    if params["family"] == "multinomial" and params["nclasses"] == 2:
        raise ValueError("Multinomial distribution with two classes not supported, use binomial distribution.")
    model = operator.raw_operator

    attr_pairs = _get_default_tree_attribute_pairs(True, params)
    fill_tree_attributes(model, attr_pairs, True)

    n_trees_in_group = params["n_trees_in_group"]
    attr_pairs['class_ids'] = [v % n_trees_in_group for v in attr_pairs['class_treeids']]
    attr_pairs['classlabels_strings'] = params["class_labels"]

    container.add_node('TreeEnsembleClassifier', operator.input_full_names,
                       operator.output_full_names,
                       op_domain='ai.onnx.ml',
                       name=scope.get_unique_operator_name('TreeEnsembleClassifier'),
                       **attr_pairs)


def convert_h2o(scope, operator, container):
    params = operator.raw_operator["params"]
    is_classifier = params["classifier"]
    if is_classifier:
        convert_classifier(scope, operator, container, params)
    else:
        convert_regression(scope, operator, container, params)


register_converter('H2OTreeMojo', convert_h2o)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._registration import register_converter


def convert_convolution(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import SamePadding

    params = operator.raw_operator.convolution
    op_type = 'ConvTranspose' if params.isDeconvolution else 'Conv'
    inputs = [operator.inputs[0].full_name]
    outputs = [operator.outputs[0].full_name]
    attrs = {'name': operator.full_name}

    n_groups = 1 if params.nGroups == 0 else params.nGroups

    shape_w = [params.outputChannels, params.kernelChannels, params.kernelSize[0], params.kernelSize[1]]
    if params.isDeconvolution:
        shape_w[0] = params.kernelChannels
        shape_w[1] = int(params.outputChannels / n_groups)
    name_w = scope.get_unique_variable_name(operator.full_name + '_W')
    inputs.append(name_w)
    container.add_initializer(name_w, onnx_proto.TensorProto.FLOAT, shape_w, params.weights.floatValue)

    if params.hasBias:
        shape_b = [len(params.bias.floatValue)]
        name_b = scope.get_unique_variable_name(operator.full_name + '_B')
        inputs.append(name_b)
        container.add_initializer(name_b, onnx_proto.TensorProto.FLOAT, shape_b, params.bias.floatValue)

    dilations = [1, 1]
    if len(params.dilationFactor) > 0:
        dilations = [params.dilationFactor[0], params.dilationFactor[1]]
    kernel_shape = [3, 3]
    if len(params.kernelSize) > 0:
        kernel_shape = params.kernelSize
    strides = [1, 1]
    if len(params.stride) > 0:
        strides = params.stride
    attrs['dilations'] = dilations
    attrs['group'] = n_groups
    attrs['kernel_shape'] = kernel_shape
    attrs['strides'] = strides

    pads = None
    auto_pad = None
    pad_type = params.WhichOneof('ConvolutionPaddingType')
    if pad_type == 'valid':
        if len(params.valid.paddingAmounts.borderAmounts) > 0:
            pads = [0, 0, 0, 0]
            pads[0] = params.valid.paddingAmounts.borderAmounts[0].startEdgeSize
            pads[1] = params.valid.paddingAmounts.borderAmounts[1].startEdgeSize
            pads[2] = params.valid.paddingAmounts.borderAmounts[0].endEdgeSize
            pads[3] = params.valid.paddingAmounts.borderAmounts[1].endEdgeSize
            # If padding amounts are all zero, there should be no padding list.
            if all(pad == 0 for pad in pads):
                pads = None
                auto_pad = 'VALID'
        else:
            auto_pad = 'VALID'
    elif pad_type == 'same':
        if params.same.asymmetryMode == SamePadding.BOTTOM_RIGHT_HEAVY:
            auto_pad = 'SAME_UPPER'
        elif params.same.asymmetryMode == SamePadding.TOP_LEFT_HEAVY:
            auto_pad = 'SAME_LOWER'
        else:
            raise ValueError('Unknown asymmetric mode: {}'.format(
                params.same.asymmetryMode))
    else:
        raise ValueError('Unsupported padding mode: {}'.format(pad_type))

    if params.isDeconvolution and len(params.outputShape) > 0:
        attrs['output_shape'] = params.outputShape

    if pads is not None:
        attrs['pads'] = pads

    if auto_pad is not None:
        attrs['auto_pad'] = auto_pad

    container.add_node(op_type, inputs, outputs, **attrs)


register_converter('convolution', convert_convolution)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_mean_variance_normalization(scope, operator, container):
    op_type = 'MeanVarianceNormalization'
    op_name = scope.get_unique_operator_name(op_type)
    params = operator.raw_operator.mvn
    attrs = {'name': op_name}
    attrs['across_channels'] = params.acrossChannels
    attrs['normalize_variance'] = params.normalizeVariance
    container.add_node(op_type, operator.input_full_names, operator.output_full_names, **attrs)


register_converter('mvn', convert_mean_variance_normalization)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_transpose
from ....common._registration import register_converter


def convert_flatten(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import FlattenLayerParams as Params

    variable_to_be_flattened_name = operator.inputs[0].full_name
    flattened_variable_name = operator.outputs[0].full_name

    if operator.raw_operator.flatten.mode == Params.CHANNEL_LAST:
        transposed_variable_name = scope.get_unique_variable_name('transposed')
        apply_transpose(scope, variable_to_be_flattened_name, transposed_variable_name, container, perm=[0, 2, 3, 1])
        variable_to_be_flattened_name = transposed_variable_name

    op_type = 'Flatten'
    flatten_attrs = {'name': operator.full_name, 'axis': 1}

    if container.target_opset < 9:
        target_opset = 1
    elif container.target_opset < 11:
        target_opset = 9
    else:
        target_opset = 11

    container.add_node(op_type, [variable_to_be_flattened_name], [flattened_variable_name], op_version=target_opset, **flatten_attrs)


register_converter('flatten', convert_flatten)
